/*
 * bibprogs.h
 *
 * Copyright (c) Chris Putnam 2008-2017
 *
 * Program and source code released under the GPL version 2
 *
 */
#ifndef BIBPROGS_H
#define BIBPROGS_H

#include "bibutils.h"

extern void bibprog( int argc, char *argv[], param *p );

#endif
